;;;; client.scm - XML-RPC client


(require 'http-client)
(require 'ssax-utils)
(require 'xml-rpc-utils)
(require 'url)
(require 'base64)

(declare 
  (uses srfi-1 extras)
  (export xml-rpc:server xml-rpc:version))


(define-constant fault-response-code 1)
(define-constant invalid-response-format-code 2)
(define-constant version "1.16")

(define xml-rpc:version 
  (sprintf "Chicken ~A, XML-RPC ~A" (chicken-version) version) )

(define (xml-rpc:server host . more)
  (define (decode-arguments host port path #!key user password (attributes '()))
    (let* ((endpoint-url (url host))
	   (endpoint-scheme (url-scheme endpoint-url))
	   (endpoint-port (url-port endpoint-url)))
      (if (member endpoint-scheme '("http" "https"))
	  (values
	   (sprintf
	    "~A://~A~A/~A"
	    endpoint-scheme
	    (url-host endpoint-url) (if endpoint-port
					(sprintf ":~A" endpoint-port)
					"")
	    (or (url-path endpoint-url) ""))
	   (or (url-user endpoint-url) user)
	   (or (url-password endpoint-url) password)
	   attributes)
	  (values
	   (sprintf
	    "http://~A~A/~A"
	    host (if port
		     (sprintf ":~A" port)
		     "")
	    path)
	   user password
	   attributes))))
  (define-values (more-optionals more-keywords)
    (break! keyword? more))
  (let-optionals more-optionals ([port #f] [path "/RPC2"])
    (let-values (((endpoint-url user password attributes)
		  (apply decode-arguments host port path more-keywords)))
      (lambda (name)
	(lambda args
	  (let ([payload (create-payload name args)])
	    (let-values ([(_ a i o)
			  (http:send-request
			   (http:make-request
			    'post 
			    endpoint-url
			    (append
			     `(("content-type" . "text/xml")
			       ("content-length" . ,(number->string (string-length payload)))
			       ("user-agent" . ,xml-rpc:version) )
			     (if (and user password)
				 `(("authorization" . ,(string-append
							"Basic "
							(base64:encode
							 (string-append
							  user ":" password)))))
				 '())
			     attributes)
			    payload
			    'http/1.0) ) ] )
			(let ([x (parameterize ([case-sensitive #t]) 
					       (SSAX:XML->SXML i '()) ) ] )
			  (close-input-port i)
			  (close-output-port o)
			  (parse-response x) ) ) ) ) ) ) ) )

(define (create-payload name args)
  (string-append
   (apply
    string-append
    "<methodCall><methodName>" name "</methodName>"
    (if (pair? args) "<params>" "")
    (map (lambda (a) (string-append "<param>" (xml-rpc:marshall-value a) "</param>")) args) )
   (if (pair? args) "</params>" "")
   "</methodCall>\n") )

(define parse-response
  (let ([bad (cut xml-rpc:error invalid-response-format-code "invalid response format" <> <>)])
    (lambda (r0)
      (match r0
	[('*TOP* items ...)
	 (let loop ([items items])
	   (match (car items)
	     [('*PI* . _) (loop (cdr items))]
	     [('|methodResponse| result)
	      (match result
		[`(params (param ,x) ...)
		 (apply values (map xml-rpc:unmarshall-value x)) ]
		[`(fault
		   (value
		    ,('struct members ...)))
		 (let loop ((code #f) (msg #f) (members members))
		   (if (null? members)
		       (xml-rpc:error (and code (string->number code))
				      (sprintf "XML-RPC fault response (code ~A): ~A" code msg))
		       (match (car members)
			 [`(member (name "faultCode") (value ,((or 'int 'i4) code)))
			  (loop code msg (cdr members))]
			 [`(member (name "faultString") (value (string ,msg)))
			  (loop code msg (cdr members))]
			 [r (bad r r0)])))]
		[r (bad r r0)] ) ]
	     [r (bad r r0)] ) ) ]
	[r (bad r r0)] ) ) ) )
