;;;; server.scm - XML-RPC server


(require 'http-server)
(require 'ssax-utils)
(require 'xml-rpc-utils)


(declare
  (export xml-rpc:register-method xml-rpc:current-method-path xml-rpc:method-documentation) )

(use extras srfi-1 srfi-69)


(define-constant invalid-parameter-format-code 3)
(define-constant invalid-rpc-format-code 4)

(http:content-parser 
 'text/xml 
 (lambda (size _ port) 
   (let ((buf (read-string size port)))
     (values 
      (SSAX:XML->SXML (open-input-string buf) '())
      buf) ) ) )

(define method-table (make-hash-table string=?))
(define rpc-url-list '())

(define xml-rpc:current-method-path (make-parameter #f))

(define-record method
  name					; string
  url					; string
  proc					; procedure
  doc)					; string

(define (lookup-method name url)
  (hash-table-ref/default method-table (string-append name " - " url) #f) )

(define (add-rpc-resource url)
  (set! rpc-url-list (cons url rpc-url-list))
  (http:add-resource
   url
   (lambda (r _)
     (let* ([data (http:request-body r)]
	    [data (if (string? data)
		      (with-input-from-string data (cut SSAX:XML->SXML (current-input-port) '())) 
		      data) ] )
       (handle-exceptions ex (write-fault-response ex r)
	 (match data
	   [(or `(*TOP* (*PI* xml . ,_) (|methodCall| ,mname . ,params))
		`(*TOP* (|methodCall| ,mname . ,params)) )
	    (match mname
	      [`(|methodName| ,name) 
	       (let* ([url (http:request-url r)]
		      [m (lookup-method name url)] )
		 (if m
		     (match params
		       [`((params (param ,x) ...))
			(receive results (call-proc m (map xml-rpc:unmarshall-value x))
			  (write-result-response results r) ) ]
		       [() 
			(receive results (call-proc m '())
			  (write-result-response results r) ) ]
		       [r (xml-rpc:error invalid-parameter-format-code "invalid parameter format" r)] )
		     (write-undefined-method-response name r) ) ) ]
	      [r (xml-rpc:error invalid-rpc-format-code "invalid rpc format" r)] ) ]
	   [r (xml-rpc:error invalid-rpc-format-code "invalid rpc format" r)] ) ) ) ) ) )

(define (call-proc m args)
  (parameterize ([xml-rpc:current-method-path (method-url m)])
    (apply (method-proc m) args) ) )

(define (xml-rpc:register-method name proc . more)
  (let-optionals more ([doc ""] [url "/RPC2"])
    (let* ([name (if (symbol? name) 
		      (symbol->string name)
		      (ensure string? name) ) ]
	   [entry (string-append name " - " url)] )
      (hash-table-set! method-table entry (make-method name url proc doc))
      (unless (member url rpc-url-list) (add-rpc-resource url)) ) ) )

(define exn? (condition-predicate 'exn))
(define exn-message (condition-property-accessor 'exn 'message))

(define (write-response s req)
  (http:write-response-header 
   200 "OK" '() (current-output-port) 
   (http:request-protocol req) )
  (printf "Content-type: text/xml\r\nContent-length: ~A\r\n\r\n~A"
	  (string-length s)
	  s) )

(define (write-fault-response ex req)
  (let ([o (open-output-string)])
    (fprintf o #<<EOF
<?xml version="1.0"?>
<methodResponse>
 <fault>
  <value>
   <struct>
    <member>
     <name>faultCode</name><value><int>~A</int></value>
    </member>
    <member>
     <name>faultString</name><value><string>~A</string></value>
    </member>
   </struct>
  </value>
 </fault>
</methodResponse>
EOF
      (if (exn? ex) 1 2)
      (if (exn? ex)
	  (exn-message ex)
	  (->string ex) ) )
    (write-response (get-output-string o) req) ) )

(define (write-undefined-method-response name req)
  (write-fault-response
   (make-property-condition 'exn 'message (sprintf "undefined method ~S" name))
   req) )

(define (write-result-response results req)
  (let ([o (open-output-string)])
    (display "<?xml version=\"1.0\"?><methodResponse>" o)
    (when (pair? results)
      (display "<params>" o)
      (for-each 
       (lambda (r) (fprintf o "<param>~A</param>" (xml-rpc:marshall-value r)))
       results)
      (display "</params>" o) )
    (display "</methodResponse>\n" o) 
    (write-response (get-output-string o) req) ) )

(define (xml-rpc:method-documentation name . url)
  (and-let* ([m (lookup-method name (:optional url "/RPC2"))])
    (method-doc m) ) )
